/* NMR spectrum calculator. Machine independent part. */
/* on PC, compile with a large data model */

/* This version compiles correctly in the compact model of either Turbo
   or Watcom C or 32 bit extended Watcom.
*/

/* new version, April 1995. In addition to interactive input, it is 
    possible to read parameters from a file of the following format:
    line 1: number of spins (N)
    line 2: P or H for ppm or Hz, followed by spectrometer freq in Mhz 
            if ppm.
    line 3: chemical shifts
    line 4: J11 J12 J13 ...
    line 5: J23 J24 ...
            ...
*/

#include "spin.h"
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>


   struct screendef scr;       /* global variables */
   FILE *hpfile;
   int ppm, nspin;

   double sf, inpfreq[MAX], j[MAX][MAX];




main(int argc, char *argv[])
{
   int i, k, kk, nt, menurow, inputrow, lorentz, Rows, recalc, dpr;
   int q1, q2, first;
   static int ntran[MAX+1] = { 0, 1, 4, 15, 56, 210, 792, 3003, 11440 T9};
   char b[81];
   double oldsf, freq[MAX], left, right, oldleft, oldright, temp, fwhm = 0.;
   struct transition huge *trans;
   FILE *output;


/* explanation of variables:
   ntran[n] contains number of transitions in general n-spin system.
   freq and j are user input frequencies and coupling constants. (Only
      upper triangle of j is actually used .)
   trans will point to  alloced space for transition freqs and intensities.
*/

   hardinit();      /* initialize hardware */

   getscreen(&scr);
   Rows = scr.crows;

   first = TRUE;

   
restart:
   ppm = ERROR;
   cursor(1,0);
   if (first && argc > 1)
   {  ppm = read_file(argv[1]);
      if (ppm == ERROR) first = FALSE;
   }
   
   if ((!first) || argc <= 1) do
   {  do
      {  prnc("\r\nEnter number of spins, or space to read from file: ");
         k = key();
      } while (k != ' ' && (k < '1' || k > '9'));

      if (k == ' ')
      {  prnc("\n\rEnter file name: ");
         zzstr(b, 80);
         ppm = read_file(b);
         if (ppm == ERROR) continue;
      }
      else
      {  nspin = k - '0';
         if (nspin < 2 || nspin > MAX)
         {  sprintf(b, "\n\rNumber of spins must be between 2 and %d\n\r", MAX);
            prnc(b);
            continue;
         }
         ppm = interactive_input(nspin);
      }
   } while (ppm == ERROR);
   
   first = FALSE;
   
assert(ppm == TRUE || ppm == FALSE);

   if (NULL == (trans = (struct transition huge *)halloc(ntran[nspin],
               sizeof(struct transition)))) errmsg("main", "Out of memory");


   lorentz = 0;
   recalc = 0;
   
   if (ppm) for (i=0;  i<nspin;  i++) freq[i] = sf * inpfreq[i];
   else for (i=0;  i<nspin;  i++) freq[i] = inpfreq[i];


recalculate:

    nmrcalc(nspin, freq, j, trans);

    if (nspin >= 5) statmsg("main", "Sorting Transitions");

    for (i=0, nt=0;  i<ntran[nspin];  i++)
    {  if (trans[i].intens < MININT) continue;  /* squeeze out very weak */
       else
       {   trans[nt] = trans[i];
           nt++;
       }
    }
   tssort(trans, nt);                           /* sort transitions */
   
   for (i=0, kk=0;  i<nt;  i++)                 /* consolidate duplicates */
   {  if (kk>0 && fabs(trans[i].de - trans[kk-1].de) < ROUND)
      {   trans[kk-1].intens += trans[i].intens;
          continue;
      }
      else
      {   trans[kk] = trans[i];
          kk++;
      }
   }
   nt = kk;
   if(ppm) for (i=0;  i<nt;  i++) trans[i].de /= sf;  /* plot will be in ppm */

    cls();

first_plot:
   if (!recalc)
   {   right= rnl(trans[0].de);        /* NMR! plot backwards */
       left = rnh(trans[nt-1].de);
       if (fabs(left - right) < 1e-10 * fabs(left + right + 1e-10))
       {   left = 2.*(left + 1e-10);
           right = right / 2.;
       }
       if (right > 0 && left/right > 5)  right = 0;
       if (ppm && right == 0 && left < 10) left = 10;
   }
   if(lorentz) lorentzplot(trans, left, right, nt, 0, fwhm);
   else stickplot(trans, left, right, nt, 0); 

   inputrow = Rows - 1;
   menurow = inputrow - 3;
   while(1)            /* loop on user action */
   {   cursor(menurow, 0);
       prnc("C:Change parm   H:HPplot     L:Lorentzians   N:New calc      O:Numeric Output");
       cursor(menurow + 1, 0);
       prnc("P:Printplot     Q:Quit       R:Replot        S:Spec.freq.");
       cursor(inputrow, 0);
       k = toupper(key());
       switch (k)
       {
           case 'C':
               cursor(inputrow, 0);
               prnc("Which parameter to change? ");
               zzstr(b, 81);
               squob(b);
               clrinp();
               if ('F' == toupper(b[0]))
               {   if (b[1] < '1' || b[1] > '0' + nspin) break;
                   q1 = b[1] - '0' - 1;
                   cursor(inputrow, 0);
                   sprintf(b, "f%d is currently %.3lf. New value ? ", q1+1,
                       inpfreq[q1]);
                   inpfreq[q1] = getparm(b);
                   if (ppm) freq[q1] = sf * inpfreq[q1];
                   else freq[q1] = inpfreq[q1];
                   recalc = 1;
                   goto recalculate;
               }
               else if ('J' == toupper(b[0]))
               {   if (b[1] < '1' || b[1] > '0' + nspin) break;
                   q1 = b[1] - '0' - 1;
                   if (b[2] <= b[1] || b[2] > '0' + nspin) break;
                   q2 = b[2] - '0' - 1;
                   cursor(inputrow, 0);
                   sprintf(b, "j%d%d is currently %.3lf. New value ? ", q1+1,
                       q2+1, j[q1][q2]);
                   j[q1][q2] = getparm(b);
                   recalc = 1;
                   goto recalculate;
               }
               else break;


           case 'H':
               cursor(inputrow, 0);
               prnc("File or device for HPGL output? ");
               zzstr(b, 81);
               hpfile = fopen(b, "a");
               if (hpfile == NULL)
               {   clrinp();
                   cursor(inputrow, 0);
                   prnc("Unable to open ");
                   prnc(b);
                   key();
                   break;
               }

               if (lorentz) lorentzplot(trans, left, right, nt, 1, fwhm);
               else stickplot(trans, left, right, nt, 1);

               fclose(hpfile);
               break;


           case 'L':
               cursor(inputrow, 0);
               fwhm = getparm("Full width at half height (in Hz) ? ");
               if (ppm) fwhm /= sf;    /* plot wil be in ppm; convert fwhm */
               cls();
               lorentz = 1;
               if (fwhm > 0) lorentzplot(trans, left, right, nt, 0, fwhm);
               else
               {   lorentz = 0;
                   stickplot(trans, left, right, nt, 0);
               }
               break;

           case 'N':
               cls();
               goto restart;


           case 'O':
               cursor(inputrow, 0);
               prnc("File or device for output? ");
               zzstr(b, 81);
               strupr(b);
               output = fopen(b, "a");
               if (output == NULL)
               {   clrinp();
                   cursor(inputrow, 0);
                   prnc("Unable to open ");
                   prnc(b);
                   key();
                   break;
               }
               fprintf(output, "\n");
               if (ppm) fprintf(output, "Spectrometer frequency: %.2lf MHz\n", sf);
               fprintf(output, "\nResonance frequencies in %s\n",
                   ppm ? "ppm" : "Hz");
               for (i=0;  i<nspin;  i++) fprintf(output, "%12.*lf\n",
                    ppm ? 3 : 2, inpfreq[i]);

               fprintf(output, "\nCoupling constants in Hz\n");
               for (i=0;  i<nspin;  i++)
               {   fprintf(output, "\n");
                   for (k=i+1;  k<nspin;  k++) fprintf(output, "%10.2lf", j[i][k]);
               }
               fprintf(output, "\n\n\n");
               if (ppm) fprintf(output, "      ppm   ");
               fprintf(output, "       Hz     Intensity\n");
               for (i=0;  i<nt;  i++)
               {   if (ppm) fprintf(output, "%11.3lf%11.2lf%12.4lf\n",
                      trans[i].de, trans[i].de * sf, trans[i].intens);
                   else fprintf(output, "%11.2lf%11.4lf\n",
                       trans[i].de, trans[i].intens);
               }
               q1 = strcmp(b, "CON");
               if (q1 == 0)
               {   fprintf(output, "\nPress any key");
                   fflush(output);
                   key();
               }
               fclose(output);
               if (q1 == 0) 
               {   cls();
                   if (lorentz) lorentzplot(trans, left, right, nt, 0, fwhm);
                   else stickplot(trans, left, right, nt, 0);
               }
               break;


           case 'P':
               cursor(inputrow, 0);
               prnc("File or device for screendump? ");
               zzstr(b, 81);

               dpr = (2 + scr.ypels) / scr.crows;     /* round for Herc. */
               i = 2 * dpr + dpr * (((3 * (scr.ypels - 1)) >> 2) / dpr);
               screendump(i, b);
               break;


           case 'Q':
               hardfinish();
               prnc("\r\nSPIN v2.0\r\n(C) copyright Ian D. Gay, 1989, 1996.\r\n"
                    "All rights reserved.\r\nEducational institutions may "
                    "use this program for teaching and research.\r\nCommercial "
                    "use prohibited.\r\n");
               exit(0);

           case 'R':
               oldleft = left;
               oldright = right;
               cursor (inputrow, 0);
               prnc("Left frequency ? ");
               zzstr(b, 81);
               if (b[0] == 'f' || b[0] == 'F') 
               {    recalc = 0;
                    cls();
                    goto first_plot;
               }                           /* expand to full range plot */

               left = atof(b);
               cursor (inputrow, 17 + strlen(b));
               right = getparm("  Right frequency ? ");
               if (left == right)
               {   left = oldleft;
                   right = oldright;
                   break;
               }
               if (left < right)
               {   temp = left;
                   left = right;
                   right = temp;
               }
               cls();
               if (lorentz) lorentzplot(trans, left, right, nt, 0, fwhm);
               else stickplot(trans, left, right, nt, 0);
               break;

           case 'S':
               if (!ppm) break;
               oldsf = sf;
               sprintf(b, "Present spectrometer frequency = %.2lf, enter new value ",
                           sf);
               cursor(inputrow, 0);
               sf = getparm(b);
               if (sf <= 0) sf = oldsf;
               fwhm *= (oldsf / sf);
               for (i=0;  i<nspin;  i++) freq[i] = sf * inpfreq[i];
               recalc = 1;
               clrinp();
               goto recalculate;

       }             /* end of switch */

       clrinp();
   }                   /* end or while(1) */
}



double getparm(char *prompt)
{
   char b[81];

   prnc(prompt);
   do { zzstr(b, 81); } while (strlen(b) <= 0);         /* reject 0 length */
   return( atof(b) );
}




void clrinp(void)
{
   cursor(scr.crows - 1, 0);
   prnc("                                                                  ");
}



#define HPBINS     1000        /* defns for HPGL output */
#define HPSTEP     9
#define HPYUNITS   6000
#define HPXOFF     500
#define HPYOFF     600
#define HPTICK     150
#define HPCDOWN    450

void hpout(char *s)        /* add s to hp output file, keeping lines <= 72 */
{
   static int len = 0;
   int l;

   l = strlen(s);
   if (len + l <= 72)
   {   fputs(s, hpfile);
       len += l;
   }
   else
   {   fputs("\n", hpfile);
       fputs(s, hpfile);
       len = l;
   }
}


void hpinit(void)          /* initialize plotter */
{  float cht;
   char b[81];

   cht = (float)HPBINS * HPSTEP / (100. * scr.ccols);
   sprintf(b, "DF;VS20;DT\006;SC250,10250,279,7479;SR%.3f,%.3f;SP1;", 
       cht, 2 * cht);
   hpout(b);
}



int axis(double left, double right, int hp) /* draw axis & return its pel row */
{
   int i, k, axpel, dpr, my, mx, trow, q, len, sigfig;
   char buf[40], b[81];
   double f, zz;



#define NTICK  5

   mx = scr.xpels - 1;
   my = scr.ypels - 1;
   dpr = (2 + scr.ypels) / scr.crows;     /* round for Herc. */

   axpel = dpr * (((3 * my) >> 2) / dpr);
   trow = 1 + axpel / dpr;

   if (hp)
   {   sprintf(b, "PU;PA%d,%d;PD;PR%d,0;PU;", HPXOFF, HPYOFF, HPSTEP * HPBINS);
       hpout(b);
   }
   else line(0, axpel, mx, axpel, GREY);

   q = axpel + (dpr >>1);
   for (i=NTICK;  i>=0;  i--) 
   {   k = -.5 + i * (mx + 1) / (float)NTICK;
       if (k > mx) k = mx;
       if (hp)
       {   sprintf(b, "PA%d,%d;PD;PR0,%d;PU;",
              HPXOFF + i * HPSTEP * (HPBINS / NTICK), HPYOFF, -HPTICK);
           hpout(b);
       }
       else line(k, axpel, k, q, GREY);
       f = left + i * (right - left) / NTICK;
       zz = fabs(.5 * (left+right) / (left - right));
       sigfig = 2.699 + ((zz < 1.0)? 0.0 : log10(zz));
       if (sigfig < 3 ) sigfig = 3;
       sprintf(buf, "%.*lg", sigfig, f);
       len = squob(buf);
       if (i == 0) k = 0;
       else if(i == NTICK) k = scr.ccols - len;
       else k = (i * scr.ccols) / NTICK - (len >> 1);
       if (hp)
       {   sprintf(b, "PA%d,%d;LB%s\006;PU;",
                HPXOFF + k * ((HPBINS * HPSTEP)/scr.ccols), HPYOFF - HPCDOWN,
                buf);
           hpout(b);
       }
       else
       {   cursor(trow, k);
           prnc(buf);
       }
   }
   if (hp)
   {   sprintf(b, "PA%d,%d;LB%s;PU;",
               (int)(HPXOFF + .48*HPBINS*HPSTEP),
                     HPYOFF - HPCDOWN, ppm ? "ppm\006" : "Hz\006");
       hpout(b);
   }
   else
   {   cursor(trow, (scr.ccols >> 1) - 1);
       prnc(ppm ? "ppm" : "Hz");
   }

   return(axpel);
}



void lorentzplot(struct transition huge *t, double left, double right,
                      int nt, int hp, double fwhm)
       /* draw plot of transitions with energies between left, right */
{   
   float *bin, maxi, yscale, b, del, reli, z, zz, n1f;
   int i, j, k, n, nbin, first, last, n1, n2, n3, k0, ax;
   char bb[81];

   if (hp) hpinit();
   ax = axis(left, right, hp);
   nbin = hp ? HPBINS : scr.xpels;

   b = 4/(fwhm * fwhm);
   del = fabs(left - right) / nbin;
   n1f = 16 * fwhm / del;
   if (n1f > 30000) return;        /* do nothing on absurdly large fwhm */
   n1 = n1f;
   n2 = n1 / 3.0;
   n3 = n2 / 3.0;

   if (NULL == (bin = (float *)malloc(nbin * sizeof(float))))
           errmsg("lorentzplot", "Out of memory");

   for (i=0;  i<nbin;  i++) bin[i] = 0;

   for (first=0;  first<nt;  first++)
               if (t[first].de >= right) break;     /* NMR! Backwards! */
   if (first >= nt || left < t[first].de)
   {   free(bin);
       return;         /* do nothing if no transitions in range */
   }

   for (last = first;  last <nt;  last++)
                if (t[last].de > left ) break;
   last--;

   maxi = 0;
   for (i=first;  i<=last;  i++)
       if (t[i].intens > maxi) maxi = t[i].intens;

   for (i = first;  i <= last;  i++)
   {   reli = t[i].intens / maxi;
       if ( reli < .001) continue;        /* We ignore lines so weak that they */
       else if (reli < .01) n = n3;       /* can't contribute .1% intensity */
       else if (reli < .1)  n = n2;       /* lorentzians are truncated at same */
       else n = n1;                       /* accuracy level */

       j = .5 + (left - t[i].de) / del;
       if (j >= nbin) j = nbin - 1;
   
       zz = z = t[i].de - (left - j * del);
       for (k = j;  (k-j) < n && k < nbin;  k++) 
       {   bin[k] += t[i].intens / ( 1 + b * z * z);
           z += del;
       }
       z = zz - del;
       for (k = j-1;  (j-k) < n && k >= 0;  k--)
       {   bin[k] += t[i].intens / (1 + b * z * z);
           z -= del;
       }
   }


   maxi = 0;
   for (i=0;  i<nbin;  i++) if (maxi < bin[i]) maxi = bin[i];
   yscale = hp ? (HPYUNITS / maxi) : ((ax - 1) / maxi);

   k0 = yscale * bin[0];
   if (hp) 
   {   sprintf(bb, "PA%d,%d;PD;PR", HPXOFF, HPYOFF + k0);
       hpout(bb);
   }

   for (i=1;  i<nbin;  i++)
   {   k = yscale * bin[i];
       if (hp)
       {   sprintf(bb, "%d,%d,", HPSTEP, k - k0);
           hpout(bb);
       }
       else line(i-1, ax-k0, i, ax-k, GREEN);
       k0 = k;
   }
   if(hp) hpout(";PU;SP;");
   free (bin);
   return;
}



void stickplot(struct transition huge *t, double left, double right,
                      int nt, int hp)
       /* draw stick plot of transitions with energies between left, right */
{
   float *bin, maxi, yscale;
   int i, k, nbin, first, last, ax, up;
   char b[81];

   if(hp) hpinit();
   ax = axis(left, right, hp);
   nbin = hp ? HPBINS : scr.xpels;

   if (NULL == (bin = (float *)malloc(nbin * sizeof(float))))
           errmsg("stickplot", "Out of memory");

   for (i=0;  i<nbin;  i++) bin[i] = 0;

   for (first=0;  first<nt;  first++)
               if (t[first].de >= right) break;     /* NMR! Backwards! */
   if (first >= nt || left < t[first].de)
   {   free(bin);
       return;         /* do nothing if no transitions in range */
   }

   for (last = first;  last <nt;  last++)
                if (t[last].de > left ) break;
   last--;

   for (i=first;  i<=last;  i++)
   {   k = -.5 + nbin * (t[i].de - left) / (right - left);
       bin[k] += t[i].intens;
   }
   maxi = 0;
   for (i=0;  i<nbin;  i++) if (maxi < bin[i]) maxi = bin[i];
   yscale = hp ? (HPYUNITS / maxi) : ((ax - 1) / maxi);

   up = 1;
   for (i=0;  i<nbin;  i++)
   {   if (bin[i] > 0)
       {   k = yscale * bin[i];
           if(hp && k > 3)
           {   if(up) sprintf(b, "PA%d,%d;PD;PR0,%d;PU;", HPXOFF + i * HPSTEP,
                   HPYOFF , k);
               else sprintf(b, "PA%d,%d;PD;PR0,%d;PU;", HPXOFF + i * HPSTEP,
                   HPYOFF + k, -k);
               hpout(b);
               up ^= 1;
           }
           else line(i, ax-1, i, ax-k, GREEN);
       }
   }
   if(hp) hpout(";PU;SP;");
   free (bin);
}



void tssort(struct transition huge *data, int n)
                 /* Shell sort, specific to struct transition */
{
    unsigned j, k, m, datastep;
    int i, mptrstep;
    struct transition huge *p1, huge *p2, temp;
 
    if (n <= 1) return;

    for (m=1;  m<=n;  m<<=1) ;  /* calculate the Hibbert step length */
    m >>= 1;
    m--;

    while (m>0)
    {   k = n - m - 1;
        mptrstep = m;
        datastep = 0;
        for (j=0;  j<=k;  j++, datastep ++)
        {   i = j;
            p1 = data + datastep;
            p2 = p1 + mptrstep;
            while ( i >= 0  && p1->de > p2->de)
            {   temp = *p1;
                *p1 = *p2;
                *p2 = temp;
                i -= m;
                p1 -= mptrstep;
                p2 -= mptrstep;
            }
        }
        m >>= 1;
    }
}



void nmrcalc( int nspin, double *freq, double j[MAX][MAX], 
                                       struct transition huge *t)
{
   static int nlast = 0;      /* holds previous value of nspin */
   static int bnc[MAX+1];     /* binomial coefficients */
   static int basis[P2MAX];   /* coded basis functions */
   static int bindex[P2MAX];  /* element i is locn of fn i in basis[] */
   static int start[MAX+1];   /* indexes to successive Fz blocks in basis[] */
   static int bits[MAX];      /* powers of 2 */

/* elements of basis code the spin product basis functions in bottom nspin
   bits. 1 bit for alpha, 0 bit for beta */

   char buf[81];
   int r, b;
   register int a, q;
   int i, k, m, f, n, bi, bk, n2, br, s2;
   int mask[MAX] = {0x1fe, 0x1fd, 0x1fb, 0x1f7, 0x1ef, 0x1df, 0x1bf, 0x17f LASTMASK};
   double /*far*/ *hspace[MAX+1];
   double /*far*/ *h[MAXBLK], far *h2[MAXBLK], *e, *e2, ti, eps, hrk, hii;

   if (nspin != nlast)        /* set up bnc, basis, etc if nspin changed */
   {  bnc[0] = 1;
      start[0] = 0;
      for(i=1, k=nspin, m=1;  i<=MAX;  i++, k--, m++)
      {  bnc[i] = (bnc[i-1] * k) / m;
         start[i] = start[i-1] + bnc[i-1];
      }
      m = 1 << nspin;
      for (i=0;  i<m;  i++)
      {  k = bitcount(i, nspin);
         bindex[i] = start[k];
         basis[(start[k]++)] = i;
      }
      start[0] = 0;
      for (i=1;  i<=MAX;  i++) start[i] = start[i-1] + bnc[i-1];

      for (i=0;  i<nspin;  i++) bits[i] = 1 << i;

      nlast = nspin;
   }

   eps = 1e100;
   for (i=0;  i<nspin;  i++)
   {  for (k=i+1;  k<nspin;  k++)
      {  if (j[i][k] == 0) continue;
         if (fabs(j[i][k]) < eps) eps = fabs(j[i][k]);
      }
   }
   if (eps == 0) eps = 1;
   eps *= ACC;             /* smallest j defines accuracy sought in energies */


   for (f=0;  f<= nspin;  f++)  /* loop over values of Fz, allocating, setting up
                                 and solving each block of the Hamiltonian. For
                                 all but first block, calculate transitions.
                                 Deallocate old blocks at bottom */
   {  n = bnc[f];
      if (NULL == (hspace[f] = /* (double far *)halloc((long)n*n, sizeof(double))*/
            (double *)malloc((unsigned)n*n*sizeof(double))))
             errmsg("nmrcalc", "Out of memory");

      if (f != 0)
      {  for (i=0;  i<bnc[f-1];  i++) h2[i] = h[i];  /*save prev row ptrs */
         e2 = e;                                     /* and energy */

         if (nspin >= 5)
         {  sprintf(buf, "Building H%d        ", f);
            statmsg("nmrcalc", buf);
         }    
      }
      if (NULL == (e = (double *)malloc(n*sizeof(double)))) 
             errmsg("nmrcalc", "Out of memory");
                                                    /* space for energies */

      for (i=0;  i<n;  i++)      /* set pointers to rows of matrix block */
         h[i] = hspace[f] + i * n;

      for (i=0;  i<n;  i++)
      {  for (k=i;  k<n;  k++)
         {  h[i][k] = 0;         /* zero the upper triangle */
         }
         bi = basis[start[f] + i];

         hii = 0;
         for (a=0;  a<nspin;  a++)
         {  if (bi & bits[a]) hii -= freq[a];
            else hii += freq[a];
         }
         hii *= .5;           /* done shift contributions to diagonal */

         for (a = 0;  a < nspin;  a++)     /* loop through couplings */
         {  for (b = a+1;  b < nspin;  b++)
            {  q = bi & bits[a];
               r = bi & bits[b];
               if ((q&&r) || (!(q||r))) hii += .25 * j[a][b];
               else hii -= .25 * j[a][b];
            }
         }                   /* done J contributions to diagonal */
         h[i][i] = hii;

         for (k=i+1;  k<n;  k++)    /* do off-diagonal terms */
         {  bk = basis[start[f] + k];
            q = bk ^ bi;      /* there is an off-diag iff basis fns differ
                                 in exactly 2 positions. The element is .5
                                 times the j for these 2 posns */
            if (bitcount(q, n) != 2) continue;
            for (m=0;  m<n;  m++)
            {  if (q & 1)
               {  a = m++;
                  q >>= 1;
                  break;
               }
               q >>= 1;
            }
            for ( ;  m<n;  m++)
            {  if (q & 1)
               {  b = m;
                  break;
               }
               q >>= 1;
            }
            h[i][k] = .5 * j[a][b];
         }                             /* off-diagonals done */
         for (k=i+1;  k<n;  k++) h[k][i] = h[i][k];

      }                                /* matrix block is complete */


      if (nspin >= 5)
      {  sprintf(buf, "Diagonalizing H%d   ", f);
         statmsg("nmrcalc", buf);
      }    


      q = ymeig(n, h[0], e, eps, 1);   /* solve the current block */

      if (q)
      {   sprintf(buf,"error %d", q);
          errmsg("ymeig", buf);
      }

/* calculate transition energies and intensities, involving this block and
   the previous one */

      if (f == 0) continue;            /* nothing on 1st block */

      if (nspin >= 5)
      {  sprintf(buf, "Transitions %d-%d    ", f-1,f);
         statmsg("nmrcalc", buf);
      }    

      n2 = bnc[f-1];
      for (i=0;  i<n2; i++)            /* i loops states in prev block */
      {  for (k=0;  k<n;  k++)         /* k loops present block */
         {  t->de = e2[i] - e[k];      /* energy done */
            s2 = start[f-1];
            ti = 0;
            for (r=0;  r<n;  r++)      /* r loops basis fns in present state */
            {  br = basis[start[f] + r];
               hrk = h[r][k];
               for (q=0;  q<nspin;  q++)   /* loop over spins */
               {   a = br & mask[q];       /* if spin q is alpha, a is the */
                   if (a != br)            /* basis fn of next lower f, to */
                   {  a = bindex[a];       /* which transition is possible */
                      ti += h2[a-s2][i] * hrk;  /* lookup tells which fn. */
if (fabs(ti) > 1000)
{  printf("ti = %.2le, i = %d, k = %d, r = %d, q = %d, hrk = %.2le, a = %d"
          "s2 = %d, h2[a-s2][i] = %.2le\n",
           ti, i, k, r, q, hrk, a, s2, h2[a-s2][i]);
   key();
}
                   }
               }
            }
            t->intens = ti * ti;
            t++;
         }
      }

      free(hspace[f-1]);            /* free dead blocks */
      free(e2);
      
      if (f == nspin)
      {  free(hspace[f]);
         free(e);
      }

   }  /* end of loop over blocks */
   return;
}



bitcount(int i,  int n)    /* count no of 1's in bottom n bits of i */
{
   int k, m;

   k = 0;
   for (m=0;  m<n;  m++)
   {  if (i & 1) k++;
      i >>= 1;
   }
   return(k);
}



ymeisb(long double p, long double q, double *c, double *s, double *n);
/* auxiliarry for ymeig */

int ymeig(n, aa, d, eps, vector)
   int n, vector;
   double *aa, *d, eps;
{
/* Transcription of Fortran routine of same name. Stewart's algorithm
   (CACM 13, 370 ( 1970), errata ibid 750) for eigenvalues, and
   optionally eigenvectors, of a real symmetric matrix.

   Arguments:
   n     dimension of matrix (must be <= 90 in 16 bit code)
   aa    input matrix. This is stored in full in a memory block of exactly
         the right size. (eg mallocked). C storage order; i.e. row-wise.
   d     output vector of length n for eigenvalues
   eps   absolute required convergence in eigenvalues
   vector if != 0, this will cause aa to be overwritten with eigenvectors

   Return value is error code, as follows:
      0: OK
     -1: Invalid n (<1 or > MAXHDIM)
     -2: Unable to malloc required workspace
    n>0: Convergence failed. n = number of initial unreliable results.
*/


   double **a, *e,  k1, ninf, norm, qmax, s2, con, temp, test;
   double cb, cc, cd, c, s, r;
   double *aj, *ai, *aLL;
   long double sum, sum1, k, k2, p, q;
   int shft, i, j, l, L1, LL, LL1, nu, num1, nl;

   double k0 = 0.2;              /* initial origin shift */
   int itlim = 50;               /* maximum allowed iterations to converge */


   if (n <= 0 || n > MAXHDIM) return(-1);

   if (n == 1)                   /* degenerate case */
   {  d[0] = aa[0];
      if (vector) aa[0] = 1;
      return(0);
   }

                                 /* allocate pointers to rows of aa */

   if (NULL == (a = (double **)malloc(n * sizeof(double *)))) return(-2);
   for (i=0;  i < n;  i++) a[i] = aa + i * n;

                                 /* allocate workspace vector */

   if (NULL == (e = (double *)malloc(n * sizeof(double))))
   {  free(a);
      return(-2);
   }


                                 /* real calculation for non-degenerate cases */

   if (n != 2)
   {  for (l=0;  l < n-2;  l++)  /* MAIN LOOP */
      {  L1 = l + 1;
         d[l] = a[l][l];
         qmax = 0;
         for (i=L1;  i < n;  i++)
         {  q = fabs(a[i][l]);
            if (qmax < q) qmax = q;
         }
         if (qmax == 0)
         {  e[l] = 0;
            a[l][l] = 1;
            continue;            /* to bottom of main loop */
         }
         sum = 0;
         for (i = L1;  i < n;  i++)
         {  ai = a[i];
            ai[l] /= qmax;
            sum += (ai[l] * ai[l]);
         }
assert(sum >= 0);
         s2 = sqrt(sum);
         if (a[L1][l] < 0) s2 = -s2;
         e[l] = -s2 * qmax;
         a[L1][l] += s2;
         a[l][l] = s2 * a[L1][l];
         sum1 = 0;
         for (i=L1;  i<n;  i++)
         {  sum = 0;
            ai = a[i];
            for (j = L1;  j <= i;  j++) sum += (ai[j]) * (a[j][l]);
            if (i != n-1) for (j = i+1;  j < n;  j++)
                          {   aj = a[j];
                              sum += aj[l] * aj[i];
                          }
            e[i] = sum / a[l][l];
            sum1 += ai[l] * e[i];    /* 50 */
         }
         con = .5 * sum1 / a[l][l];
         for (i = L1;  i < n;  i++)
         {  e[i] -= con * a[i][l];
            ai = a[i];
            for (j = L1;  j <= i;  j++)
                     ai[j] -= (ai[l] * e[j]  +  a[j][l] * e[i]);
         }
      }                          /* 70 */
   }                             /* end of if ( n != 2) */

   d[n-2] = a[n-2][n-2];
   d[n-1] = a[n-1][n-1];
   e[n-2] = a[n-1][n-2];

   if (vector)
   {  a[n-2][n-2] = a[n-1][n-1] = 1;
      a[n-2][n-1] = a[n-1][n-2] = 0;

      if (n != 2)
      {  for (LL = n-3;  LL >= 0;  LL--)
         {  LL1 = LL + 1;
            aLL = a[LL];
            for (i = LL1;  i < n;  i++)
            {  sum = 0;
               for (j = LL1;  j < n;  j++)  sum += a[j][LL] * a[j][i];
               aLL[i] = sum / aLL[LL];
            }                    /* 140 */

            for (i = LL1;  i < n;  i++)
            {  ai = a[i];
               for (j = LL1;  j < n;  j++) ai[j] -= ai[LL] * aLL[j];
            }
            for (i = LL1;  i < n;  i++) a[i][LL] = aLL[i] = 0;
            aLL[LL] = 1;
         }                       /* 170 */
      }
   }

   ninf = (fabs(d[0]) + fabs(e[0]) > fabs(d[n-1]) + fabs(e[n-2])) ?
               (fabs(d[0]) + fabs(e[0])) : ( fabs(d[n-1]) + fabs(e[n-2]));

   if (n != 2) for (i = 1;  i < n-1;  i++)
               {  q = fabs(d[i]) + fabs(e[i]) + fabs(e[i-1]);
                  if (ninf < q) ninf = q;
               }

   nu = n - 1;
   num1 = nu - 1;
   shft = 0;
   k1 = k0;
   test = ninf * eps;
   e[n-1] = 0;

   while(1)             /* iterate to convergence or failure */
   {  for (nl = num1;  nl >= 0;  nl--)
      {  if (fabs(e[nl]) <= test ) break;
      }
      if (nl >= 0)
      {  e[nl++] = 0;            /* 230 */
         if (nl == nu)
         {  if (num1 == 0)
            {  free(a);
               free(e);
               return(0);
            }
            nu = num1;
            num1 = nu - 1;
            continue;            /* rerun while(1) loop */
         }
      }
      nl = 0;                    /* 240 */
      e[nu] += 1;
      if (e[nu] > itlim)
      {  free(a);
         free(e);
         return(nu);
      }

      cb = (d[num1] - d[nu]) / 2;   /* 250 */
      qmax = (fabs(cb) > fabs(e[num1])) ? fabs(cb) : fabs(e[num1]);
assert(qmax != 0);
      cb /= qmax;
      cc = e[num1] / qmax;
      cc *= cc;
assert(cb * cb + cc >= 0);
      cd = sqrt(cb * cb + cc);
      if(cb != 0) cd = fabs(cd) * ((cb > 0) ? 1 : -1);
assert(cb + cd != 0);
      k2 = d[nu] - qmax * cc / (cb + cd);

      if (shft) k = k2;
      else if (fabs(k2 - k1) < .5 * fabs(k2))
      {  shft = 1;
         k = k2;
      }
      else
      {  k1 = k2;
         k = k0;
      }

      ymeisb(d[nl] - k, e[nl], &c, &s, &norm);     /* 300 */

      for (i = nl;  i <= num1;  i++)
      {  if(vector)
         {  for (j=0;  j < n;  j++)
            {  aj = a[j];
               temp = c * aj[i] + s * aj[i+1];
               aj[i+1] = -s * aj[i] + c * aj[i+1];
               aj[i] = temp;
            }
         }
         d[i] = c * d[i] + s * e[i];
         temp = c * e[i] + s * d[i+1];
         d[i+1] = -s * e[i] + c * d[i+1];
         e[i] = -s * k;
         d[i] = c * d[i] + s * temp;
         if (i == num1) continue;
         if (fabs(s) <= fabs(c))
	      {  assert(c != 0);
            r = s / c;
            d[i+1] = -s * e[i] + c * d[i+1];
            e[i+1] *= c;
            ymeisb(d[i+1] - k, e[i+1], &c, &s, &norm);
            e[i] = r * norm;
         }
         else
         {  p = c * e[i] + s * d[i+1];
            q = s * e[i+1];
assert(s != 0);
            d[i+1] = k + c * p / s;
            e[i+1] *= c;
            ymeisb(p, q, &c, &s, e+i);
         }
      }                                /* 380 */

      temp = c * e[num1] + s * d[nu];
      d[nu] = -s * e[num1] + c * d[nu];
      e[num1] = temp;
   }                                   /* end of while(1) iteration */
}


ymeisb(long double p, long double q, double *c, double *s, double *n)
{  
   if (fabs(q) >= fabs(p))    /* original has > but that doesnt make sense */
   {  if (q == 0)
      {  *c = 1;
         (*s) = (*n) = 0;
      }
      else
      {  assert(q != 0);
         *n = fabs(q) * sqrt(1 + p * p / (q * q));
   assert(*n != 0);
         (*c) = p / (*n);
         (*s) = q / (*n);
      }
   }
   else
   {  if(p == 0) (*n) = q;
/* limiting form of following statement, which crashes for p = 0. Should
   this be possible? */
      else (*n) = fabs(p) * sqrt(1 + q * q / (p * p));

   assert((*n) != 0);  
      (*c) = p / (*n);
      (*s) = q / (*n);
   }
}



int squob(char *s)  /* squeeze out the blanks in s and return the length */
                    /* of the resultant string not counting terminating null*/
{   int i,j;
    char c;

    i=j=0;
    while ( (c=s[i++]) != '\0')     if (c != ' ') s[j++]=c;
    s[j]='\0';
    return(j);
}


int read_file(char *n)
{  char b[81], bb[81], *p;
   FILE *input;
   int i, k, ppm;
   
   input = fopen(n, "rt");
   if (input == NULL)
   {  sprintf(b, "\r\nCan\'t open %s", n);
      prnc(b);
      return(ERROR);
   }
   if (EOF == qgets(b, 81, input)) return(ERROR);
   nspin = atoi(b);
   if (nspin < 2 || nspin > MAX) goto bad_input;
   
   if (EOF == qgets(b, 81, input)) return(ERROR);
   strupr(b);
   p = strpbrk(b, "PH");
   if (p == NULL) goto bad_input;
   if (*p == 'P') ppm = TRUE;
   else ppm = FALSE;
   if (ppm) sf = atof(p + 1);
      
   if (EOF == qgets(b, 81, input)) return(ERROR);
   p = b;
   for (i=0;  i<nspin;  i++) 
   {  if (*p == 0) goto bad_input;
      inpfreq[i] = strtod(p, &p);
	}
   for (i=0;  i<nspin - 1;  i++)
	{  if (EOF == qgets(b, 81, input)) return (ERROR);
	   p = b;
	   for (k=i+1;  k<nspin;  k++)
	   {  if (*p == 0) goto bad_input;
	      j[i][k] = strtod(p, &p);
	   }
   }
   return(ppm);

bad_input:
   sprintf(bb, "Bad input: %s\n\r", b);
   prnc(bb);
   return(ERROR);
}



int qgets(char *b, int size, FILE *f)  /* like fgets but eliminates comments */
{  char *p;                            /* starting with #. returns EOF on */
   int i, n;                           /* EOF, TRUE otherwise */

   do
   {  p = fgets(b, size, f);
      if (p == NULL) return(EOF);
      
      p = strchr(b, '#');
      if (p != NULL) *p = 0;
      n = strlen(b);
      for (i=n-1;  i>=0;  i--)
      {  if (b[i] == ' ' || b[i] == '\n' ) b[i] = 0; 
                                 /* strip trailing blanks or newlines */
         else break;
      }
   } while (strlen(b) == 0);

   return(TRUE);
}



double rnl(double x)       /* round number below x */
{
   return(-rnh(-x));
}


double rnh(double x)       /* get round number above x */
{
    double p, t, r, ax;
    int j;

   ax = x;
   if (ax < 0) ax = -ax;    
    if (ax < 1e-10) return(0);

    p = log10(ax);
    j=(int)((p < 0)? p-1 : p); 
    t=pow(10.0, j);
    r=x/t;
    j=(r>=0)?  r+.999999 : r;

    if (j==2 && r<=1.5) return(1.5*t);
    if (j==-1 && r<=-1.5) return(-1.5*t);
    if (j==7) j=8;
    if (j==9) j=10;
    if (j==-7) j=-6;
    if (j==-9) j=-8;
    return(j*t);
}



int interactive_input(int nsp)
{  char b[81];
   int i, k;

   prnc("\n\n\rFrequencies in Hz or ppm? (H/P) ");
   if ('P' == toupper(key())) ppm = TRUE;
   else ppm = FALSE;

   prnc("\n\n\r");
   
   if(ppm) do
   {  sf = getparm("Enter spectrometer frequency in MHz ");
      prnc("\n\n\r");
   } while(sf <= 0);
   else prnc("\n\n\r");

   if (ppm) prnc("Enter resonance frequencies in ppm\n\n\r");
   else  prnc("Enter resonance frequencies in Hz\n\n\r");

   for (i=0;  i<nsp; i++)
   {  sprintf(b, "   f%d = ", i+1);
      inpfreq[i] = getparm(b);
   }
   
   prnc("\n\n\rEnter coupling constants in Hz\n\n\r");
   
   for (i=0;  i<nsp;  i++)
   {  for (k=i+1;  k<nsp; k++)
      {  sprintf(b, "   j%d%d = ", i+1, k+1);
         j[i][k] = getparm(b);
      }
      prnc("\n\r");
   }
   return(ppm);
}
